import pyspark.sql.types as T
from pyspark.sql import Window
from pyspark.sql import functions as F

import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer

SEQUENCE_HASH = "sequenceHash"


class DistinctSequenceTransformer(LoggableTransformer):
    """
    Deduplicate a dataframe by a column containing a collection. This can e.g. be used to remove duplicate
    playlists from a dataframe. Order is ignored meaning that the lists [1, 2, 3] and [2, 1, 3] will be considered
    duplicates. For draws we randomly select a row and discard all others.

    :param sequence_col: column containing the sequence
    """
    def __init__(self, sequence_col: str):
        super().__init__()
        self.sequence_col = sequence_col

    def _transform(self, dataset):
        hashes = dataset.withColumn(SEQUENCE_HASH, sequence_hash(F.col(self.sequence_col)))
        window = Window.partitionBy(SEQUENCE_HASH).orderBy(F.rand())
        return (hashes
                .withColumn(c.RN, F.row_number().over(window))
                .where(F.col(c.RN) == 1)
                .drop(SEQUENCE_HASH, c.RN))


@F.udf(returnType=T.StringType())
def sequence_hash(sequence):
    return hash(tuple(sorted(sequence)))



from typing import Any

from pyspark.sql import DataFrame
from tidal_per_transformers.transformers import LoggableTransformer


class JoinTransformer(LoggableTransformer):
    """Transformer to join to spark dataframes

    :param df: DataFrame to join with
    :param on: Join condition
    :param how: Join type
    """

    def __init__(self, df: DataFrame, on: Any, how: str = "inner", hint: str = None):
        super().__init__()
        self.df = df.hint(hint, on) if hint else df
        self.join_condition = on
        self.join_type = how

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return dataset.join(self.df, self.join_condition, self.join_type)



import pyspark.sql.functions as F
from pyspark.sql.window import Window
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer


class TopItemsTransformer(LoggableTransformer):
    """
    Take the top 'n' records from a partition ordered by 'order_by'
    :param partition:   The columns used to partition the DataFrame
    :param order_by:    The column(s) used to sort the partitions
    :param n:           Number of top items to keep per partition
    :param ranking_function: Ranking function, e.g. "F.dense_rank" to account for ties. Default is ranking by rows, ignoring ties.
    :param keep_rnk_col: Option to keep or drop rank column

    """
    def __init__(self, partition, order_by, n, ranking_function=F.row_number, keep_rnk_col=False):
        super(TopItemsTransformer, self).__init__()
        self.partition = partition
        self.order_by = order_by
        self.n = n
        self.ranking_function = ranking_function
        self.keep_rnk_col = keep_rnk_col

    def _transform(self, dataset):
        window = Window.partitionBy(self.partition).orderBy(self.order_by)
        output = dataset.withColumn("rn", self.ranking_function().over(window)).where(F.col("rn") <= self.n)
        if not self.keep_rnk_col:
            output = output.drop("rn")
        return output



from pyspark.sql import functions as F, DataFrame
from tidal_per_transformers.transformers import LoggableTransformer

import tidal_per_transformers.transformers.utils.constants as c


class ArtistCompoundMappingTransformer(LoggableTransformer):
    """
    Map the artist compound id's (e.g. Miguel feat. Travis Scott -> Miguel, Travis Scott)

    TODO: Find a better way of picking the top n featuring artist per compound id

    :returns: DataFrame where the compound id's each have been mapped to a separate row. E.g. when using this on the
              playback log you will get one log entry per artist.
    """
    def __init__(self, max_featuring_artists=3, keep_info=False, compound_artist_mapping: DataFrame = None):
        super(ArtistCompoundMappingTransformer, self).__init__()
        self.max_featuring_artists = max_featuring_artists
        self.keep_info = keep_info
        self.compound_artist_mapping = compound_artist_mapping

    def _transform(self, dataset):
        compound_map = (self.compound_artist_mapping
                        .withColumnRenamed(c.ARTIST_COMPOUND_ID, c.RESOLVED_ARTIST_ID)
                        .drop(c.ID))

        # If there is no compound entry we already have the main artist
        joined = (dataset
                  .join(compound_map, c.ARTIST_ID, "left")
                  .withColumn(c.PRIORITY, F.when(F.col(c.PRIORITY).isNull(), F.lit(0)).otherwise(F.col(c.PRIORITY)))
                  .where(F.col(c.PRIORITY) <= self.max_featuring_artists))

        mapped = (joined
                  .withColumn(c.ARTIST_ID, F.when(F.col(c.RESOLVED_ARTIST_ID).isNotNull(),
                                                  F.col(c.RESOLVED_ARTIST_ID)).otherwise(F.col(c.ARTIST_ID)))
                  .withColumn(c.MAIN_ARTIST, F.when(F.col(c.MAIN_ARTIST) == "true", F.lit(1)).otherwise(F.lit(0)))
            .drop(c.RESOLVED_ARTIST_ID, c.ARTIST_COMPOUND_ID))

        if not self.keep_info:
            mapped = mapped.drop(c.PRIORITY, c.MAIN_ARTIST)

        return mapped



from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer
import pyspark.sql.types as T
import pyspark.sql.functions as F


class DuplicateSequenceTransformer(LoggableTransformer):
    """ Remove duplicate items from a sequence """
    def __init__(self, column: str, id_field: str, schema: T.ArrayType):
        super().__init__()
        self.column = column
        self.id_field = id_field
        self.schema = schema

    def _transform(self, dataset):
        @F.udf(returnType=self.schema)
        def duplicate_sequence(sequence, id_field):
            res = []
            seen = set()
            for item in sequence:
                if item[id_field] not in seen:
                    res.append(item)
                seen.add(item[id_field])
            return res

        return dataset.withColumn(self.column, duplicate_sequence(self.column, F.lit(self.id_field)))



from pyspark.sql import Window, DataFrame, functions as F
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer
import tidal_per_transformers.transformers.utils.constants as c


class PlaylistCoherenceTransformer(LoggableTransformer):
    """
    Filter playlists based on how coherence according to the playlist_coherence model

    threshold: minimum coherence value to keep the playlist
    playlist_coherence_model: model used to compute the coherence
    """
    PERCENTILE = "percentile"

    def __init__(self, threshold: float, playlist_coherence_model: DataFrame):
        super().__init__()
        self.threshold = threshold
        self.model = playlist_coherence_model
        self.model = self.model.withColumn(c.ID, F.col(c.PLAYLIST_ID)) if c.ID not in self.model.columns else self.model

    def _transform(self, dataset):
        dataset = dataset.withColumn(c.ID, F.col(c.PLAYLIST_ID)) if c.ID not in dataset.columns else dataset

        non_coherent = (self.model
                        .withColumn(self.PERCENTILE, F.percent_rank().over(Window.orderBy(c.SCORE)))
                        .where(F.col(self.PERCENTILE) > self.threshold)
                        .select(F.col(c.ID)))

        return dataset.join(non_coherent, c.ID, "left_anti")



import pyspark.sql.functions as F
import tidal_per_transformers.transformers.utils.constants as c

from pyspark.sql.window import Window
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer
from tidal_per_transformers.transformers.utils.dither_utils import dither


class TopItemsDitheringTransformer(LoggableTransformer):
    """
    Take the top 'n' records from a partition ordered by 'order_by' after applying dithering on the original
    relevance ordering.
    """
    def __init__(self, partition, order_by, n, epsilon=2.0, start_pos=0, keep_rank_col=False):
        super(TopItemsDitheringTransformer, self).__init__()
        self.partition = partition
        self.order_by = order_by
        self.n = n
        self.epsilon = epsilon
        self.start_pos = start_pos
        self.keep_rank_col = keep_rank_col

    def _transform(self, dataset):
        window = Window.partitionBy(self.partition).orderBy(self.order_by)
        ranked = dataset.withColumn(c.RN, F.row_number().over(window))
        dithered = dither(ranked, c.RN, self.partition, self.epsilon, self.start_pos).where(f"{c.RN} <= {self.n}")

        if self.keep_rank_col:
            return dithered
        else:
            return dithered.drop(c.RN)



from typing import Union, Dict

from pyspark.sql import DataFrame

from tidal_per_transformers.transformers import LoggableTransformer


class ReplaceNullValuesTransformer(LoggableTransformer):
    """Transformer to replace null values
    """
    def __init__(self, value: Union[object, Dict]):
        super().__init__()
        self.value = value

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return dataset.na.fill(self.value)



import pyspark.sql.functions as F
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer
import tidal_per_transformers.transformers.utils.constants as c


class MinMaxScalingTransformer(LoggableTransformer):
    """
    Min max transformation
    """
    def __init__(self, input_score_range=None, target_score_range=(1, 5), rounding=True):
        super().__init__()
        if input_score_range is not None:
            assert input_score_range[1] > input_score_range[0], "invalid input score range"

        self.input_score_range = input_score_range
        self.target_score_range = target_score_range
        self.rounding = rounding

    def _transform(self, scores):
        if self.input_score_range is None:
            min_max_score = scores.select(F.min(c.SCORE), F.max(c.SCORE)).collect()[0]
            self.input_score_range = min_max_score[0], min_max_score[1]

        input_range = self.input_score_range[1] - self.input_score_range[0]
        output_range = self.target_score_range[1] - self.target_score_range[0]
        scores = (scores
                  .withColumn(c.SCORE, (F.col(c.SCORE) - F.lit(self.input_score_range[0])) / F.lit(input_range))
                  .withColumn(c.SCORE, F.col(c.SCORE) * output_range + self.target_score_range[0]))

        if self.rounding:
            scores = scores.withColumn(c.SCORE, F.round(c.SCORE))
        return scores



import pyspark.sql.functions as F
import pyspark.sql.types as T
from pyspark.sql.dataframe import DataFrame

import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers import ArtistFilterTransformer


class ArtistsStructFilterTransformer(ArtistFilterTransformer):
    """
    :param artist_filters:          DataFrame containing the content filter information
    :param remove_ambient_music:    flag for toggling ambient music on/off
    :param remove_holiday_music:    flag for toggling holiday music on/off
    :param remove_children_music:   flag for toggling children music on/off
    :param min_artist_streams:      minimum number of streams for an artist to be included
    :param min_artist_streamers:    minimum number of unique listeners for an artist to be included
    :param artist_column:           Column containing artists
    """

    def __init__(self,
                 artist_filters: DataFrame,
                 remove_children_music: bool,
                 remove_ambient_music: bool,
                 remove_holiday_music: bool,
                 min_artist_streamers: int = 500,
                 min_artist_streams: int = 2000,
                 artist_column: str = c.ARTISTS
                 ):
        super().__init__(artist_filters,
                         remove_children_music,
                         remove_ambient_music,
                         remove_holiday_music,
                         min_artist_streamers,
                         min_artist_streams)
        self.artist_filters = artist_filters
        self.remove_children_music = remove_children_music
        self.remove_ambient_music = remove_ambient_music
        self.remove_holiday_music = remove_holiday_music
        self.min_artist_streamers = min_artist_streamers
        self.min_artist_streams = min_artist_streams
        self.artist_column = artist_column

    def _transform(self, dataset):
        cleaned_artists = self.apply_filters(self.artist_filters,
                                             self.min_artist_streams,
                                             self.min_artist_streamers,
                                             self.remove_holiday_music,
                                             self.remove_ambient_music,
                                             self.remove_children_music)
        artists_set = set(cleaned_artists.toPandas()[c.ARTIST_ID].tolist())
        artists_set_bc = dataset.sql_ctx.sparkSession.sparkContext.broadcast(artists_set)

        @F.udf(returnType=T.ArrayType(T.IntegerType()))
        def filter_artists(main_artists):
            return [artist[c.ID] for artist in main_artists if artist[c.ID] in artists_set_bc.value]

        return (dataset
                .withColumn(c.MATCHES, filter_artists(self.artist_column))
                .where(F.size(c.MATCHES) > 0)
                .drop(c.MATCHES))



from typing import Optional

import pyspark.sql.functions as F

from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer


class BlazingTextInputFormatTransformer(LoggableTransformer):
    """ Convert an array column to the input format required by BlazingText """

    def __init__(self, items_column: str, alias: Optional[str] = None):
        """
        :param items_column: column containing the input sequences
        :param alias:        name given to the output column (optional)
        """
        super(BlazingTextInputFormatTransformer, self).__init__()
        self.items_column = items_column
        self.alias = alias if alias else items_column

    def _transform(self, dataset):
        return dataset.select(F.array_join(F.col(self.items_column), " ").alias(self.alias))




from pyspark.sql.dataframe import DataFrame

from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer


class UnionTransformer(LoggableTransformer):
    def __init__(self, other: DataFrame):
        super().__init__()
        self.other = other

    def _transform(self, dataset):
        return dataset.union(self.other.select(dataset.columns))



from enum import Enum


# ----- Enums -----
class Artifact(Enum):
    TRACK = "track"
    TRACK_GROUP = "trackGroup"
    ARTIST = "artist"
    TRACKBUNDLE = "trackBundle"
    MASTER_BUNDLE_ID = "masterBundle"
    MASTERBUNDLE = "masterBundle"
    PLAYLIST = "playlist"
    VIDEO = "video"

    
class PlaylistCategory(Enum):
    DOLBY = "DOLBY"
    IMMERSIVE = "IMMERSIVE"
    RISING = "RISING"
    PRODUCED_BY = "PRODUCED_BY"
    FLAGSHIPS = "FLAGSHIPS"
    CLASSICS = "CLASSICS"
    BEST_OF = "BEST_OF"
    ESSENTIALS = "ESSENTIALS"
    TRACKTRACES = "TRACKTRACES"
    HITS = "HITS"
    ONE_O_ONE = "101"

    
MAIN_ARTIST_FILTER = "filter(artists, x -> x.main == 'true')"

RN = "rn"
ITEM = "item"
ITEMS = "items"
ITEMS_TO_KEEP = "itemsToKeep"
USER = "user"
SCORE = "score"
TRACK_GROUP = "trackGroup"
COUNT = "count"
HIDDEN_PARAMS = "_params"
DELETED = "deleted"
AVAILABLE = "available"
AVAILABLE_COUNTRY_CODES = "AvailableCountryCodes"
SIZE = "size"

CONTRIBUTOR_ID = "contributorId"
CONTRIBUTOR_TYPE = "contributorType"


# ----- Dataset Constants -----
# IDs
ID = "id"
UID = "uid"
USER_ID = "userId"
ARTIST_ID = "artistId"
MAIN_ARTIST_ID = "mainArtistId"
PLAYLIST_ID = "playlistId"
PRODUCT_ID = "productId"
SOURCE_ID = "sourceId"
ALBUM_ID = "albumId"
MASTER_BUNDLE_ID = "masterBundleId"
TRACK_BUNDLE_ID = "trackBundleId"
MIX_ID = "mixId"
ARTIFACT_ID = "artifact_id"
VIDEO_ID = "videoId"


DATA_FORMAT = "dataFormat"
TRACKS = "tracks"
ARTISTS = "artists"
VOICE = "voice"
PRODUCT_TYPE = "productType"
ACTUAL_PRODUCT_ID = "actualProductId"
SOURCE_TYPE = "sourceType"
START_TIMESTAMP = "startTimestamp"
END_TIMESTAMP = "endTimestamp"
STREAM_DURATION = "streamDuration"
SERVER_TS = "serverTs"
COUNTRY_CODES = "countryCodes"
COUNTRY_CODE = "countryCode"
TITLE = "title"

AUDIO_QUALITY = "audioQuality"
GENRE = "genre"
SUM = "sum"
TYPE = "type"
VALUE = "value"
VARIABLE = "variable"
DT = 'dt'

# ----- Rovi dataset constants -----
ROVI = "ROVI"
SOURCE = "source"
ROVI_SOURCE_ID = "source_id"
ROVI_ARTIFACT_TYPE_ID = "artifact_type_id"
ROVI_ARTIFACT_ID = "artifact_id"
ROVI_METADATA_ID = "metadata_id"
ROVI_ARTIST = "ARTIST"
METADATA_TYPE_ID = "metadata_type_id"
PLAYBACKLOG = "playbacklog"
USER_STATE = "userstate"


ROVI_ACTIVE_YEARS = "Active Years"
ACTIVE_YEARS = "activeYears"
ROVI_COUNTRY_OF_ORIGIN = "Country of Origin"
COUNTRY_OF_ORIGINS = "countryOfOrigins"
ROVI_INSTRUMENTATION_ENSEMBLE = "Instrumentation-Ensemble"
INSTRUMENTATION_ENSEMBLES = "instrumentationEnsembles"
ROVI_NAME_ROLE = "Name Role"
NAME_ROLES = "nameRoles"
ROVI_TONE = "Tone"
TONES = "tones"
ROVI_THEME = "Theme"
THEMES = "themes"
ROVI_ORIGINAL_LANGUAGE = "Original language"
ORIGINAL_LANGUAGE = "originalLanguage"
ROVI_GROUP_MEMBER_CONTRIBUTOR_DATES = "Group Member/Contributor Dates"
GROUP_MEMBER_CONTRIBUTOR_DATES = "groupMemberContributorDates"
ROVI_GENDER = "Gender"
GENDERS = "genders"
ROVI_MUSICAL_GENRE = "Musical Genre"
MUSICAL_GENRES = "musicalGenres"
ROVI_MAIN_CREDIT_INSTRUMENT = "Main Credit/Instrument"
MAIN_CREDIT_INSTRUMENTS = "mainCreditInstruments"
ROVI_PERIOD = "Period"
PERIODS = "periods"
ROVI_SHORT_NOTE = "Short Note"
SHORT_NOTES = "shortNotes"
ROVI_DECADE = "Decade"
DECADES = "decades"
ROVI_AREA_OF_OPERATION = "Area Of Operation"
AREA_OF_OPERATION = "areaOfOperation"
ROVI_ENSEMBLE_TYPE = "Ensemble Type"
ENSEMBLE_TYPE = "ensembleType"
ROVI_BIRTH_DATE = "Birth Date"
BIRTH_DATES = "birthDates"
ROVI_BIRTH_NAME = "Birth Name"
BIRTH_NAMES = "birthNames"
ROVI_DEATH_DATE = "Death Date"
DEATH_DATES = "deathDates"

ASSOCIATION_NAME = "association_name"

START = "start"
END = "end"

ALIASES = "aliases"
PICTURE = "picture"

STEPPED_LENGTH = "steppedLength"
PARTITION = "partition"
GROUP = "group"
INDEX = "index"


class MetadataType(Enum):
    BIOGRAPHY = 1
    ASSOCIATION = 2
    ARTIFACT_IMAGE = 3
    REVIEW = 4
    LINK = 5
    TOP_LISTS = 7
    ATTRIBUTE = 8
    PRODUCT_IMAGE = 9
    GENRE = 10
    CREDIT = 11


class AssociationName(Enum):
    ARTIST_HAS_PERFORMED_SONGS_BY = "ARTIST_HAS_PERFORMED_SONGS_BY"
    PRIMARY_NAME = "PRIMARY_NAME"
    ARTIST_COMPOSER_NAME = "ARTIST_COMPOSER_NAME"
    CONDUCTED = "CONDUCTED"
    RELATIVE_OF = "RELATIVE_OF"
    ARTIST_SIMILAR_TO = "ARTIST_SIMILAR_TO"
    SIBLING_OF = "SIBLING_OF"
    CHILD_OF = "CHILD_OF"
    STUDIED_WITH = "STUDIED_WITH"
    ALSO_PERFORMED_AS = "ALSO_PERFORMED_AS"
    ARTIST_PERFORMED_WORKS_BY = "ARTIST_PERFORMED_WORKS_BY"
    COLLABORATOR_WITH = "COLLABORATOR_WITH"
    FOUNDER = "FOUNDER"
    FOLLOWED = "FOLLOWED"
    INFLUENCED = "INFLUENCED"
    PARENT_OF = "PARENT_OF"
    ARTIST_WORKS_PERFORMED_BY = "ARTIST_WORKS_PERFORMED_BY"
    MEMBER_OF = "MEMBER_OF"
    IS_ASSOCIATED_WITH = "IS_ASSOCIATED_WITH"
    MARRIED_TO = "MARRIED_TO"
    LABEL_GROUP_FOR = "LABEL_GROUP_FOR"
    COMPOSER_SIMILAR_TO = "COMPOSER_SIMILAR_TO"
    GRANDPARENT_OF = "GRANDPARENT_OF"
    GRANDCHILD_OF = "GRANDCHILD_OF"
    DISTRIBUTOR = "DISTRIBUTOR"
    CONTRIBUTOR_FOR = "CONTRIBUTOR_FOR"



# ----- Column Names -----
DATE_ADDED = "dateAdded"
TRACK_ID = "trackId"
FAVORITE = "favorite"
PROFILE_ID = "profileId"
LAST_UPDATED = "lastUpdated"
REGISTERED_DATE = "registeredDate"
LAST_DATE_PROCESSED = "lastDateProcessed"
POPULARITY = "popularity"
POPULARITY_WW = "popularityWW"
POPULARITY_WEIGHTED = "popularityWeighted"
POPULARITY_ALL_TIME = "popularityAllTime"
ACTUAL = "actual"
VARIOUS_ARTIST_ALBUM = "variousArtistAlbum"
ALBUM = "album"
ALBUM_TYPE = "albumType"
ARTIST_IMAGE = "artistImage"
RELEASE_DATE = "releaseDate"
NUM_FEATURING_ALBUMS = "numFeaturingAlbums"
MIN_RELEASE_DATE = "minReleaseDate"
LAST_STREAMED_DATE = "lastStreamedDate"


# ---- Misc Constants -----
MAIN_ARTISTS = "mainArtists"
MAIN_ARTIST = "mainArtist"
FEATURING_ARTISTS = "featuringArtists"
NAME = "name"
AGE = "age"
ARTIST_NAME = "artistName"
ARTIST_AGE = "artistAge"
MAIN = "main"
NON_MUSIC = "non_music"
STREAM_COUNT = "streamCount"
STREAMERS_COUNT = "streamersCount"
DURATION = "duration"
LOG_ENTRY_DATE = "logEntryDate"
COMPLETION = "completion"
DATETIME = "datetime"
CHILDREN = "children"
AMBIENT = "ambient"
HOLIDAY = "holiday"
MATCHES = "matches"
ARTIST_COMPOUND_ID = "artistCompoundId"
RESOLVED_ARTIST_ID = "resolvedArtistId"
PRIORITY = "priority"
EXISTING = "existing"
UPDATED = "updated"
DATE = "date"
RESULT = 'result'
EXCEPTION_CONFIG = 'expectation_config'
DAY = "day"
MONTH = "month"
YEAR = "year"
MASTER_ALBUM = "masterAlbum"
DATABRICKS_MLFLOW_URI = "databricks"

# ----- Feature store -----

TRANSACTIONS = "transactions"
STREAMS = "STREAMS"
USER_TRACK = "user_track"
USER_ARTIST = "user_artist"
USER_ARTIST_STREAMS = "user_artist_streams"
USER_ARTISTS_DISCOVERY_OBSERVED = "UserArtistsDiscoveryObserved"
USER_ARTISTS_FAVOURITES = 'UserArtistsFavourite'
USER_ARTISTS_HISTORY = 'UserArtistsHistory'
USER_ARTISTS_PLAYLISTS = 'UserArtistsPlayLists'
USER_TRACKS_DISCOVERY_OBSERVED = 'UserTracksDiscoveryObserved'
USER_TRACKS_FAVOURITES = 'UserTracksFavourite'
USER_TRACKS_HISTORY = 'UserTracksHistory'
USER_TRACKS_PLAYLIST = 'UserTracksPlayLists'
NUM_MAIN_ARTISTS = "numMainArtists"
ARTIFACT_TYPE = "artifactType"
COMPOUND_ID = "compoundId"
RANK = "rank"

# ----- Celeste Metadata -----
METADATA_ID = "metadata_id"
METADATA_TYPE = "metadataType"
METADATA_VALUE = "value"
ASSOCIATION_ROLE_ID = "association_role_id"
ASSOCIATION_ARTIFACT_TYPE_ID = "association_artifact_type_id"
ASSOCIATION_ARTIFACT_ID = "association_artifact_id"
ASSOCIATION_ID = "associationId"
WEIGHT = "weight"


# ----- FS -----
METADATA = 'metadata'
USERS = 'users'

ARTIFACT_TYPE_ID = {
    "TRACK": 1,
    "ARTIST": 2,
    "TRACKBUNDLE": 3,
    "VIDEO": 4,
    "PLAYLIST": 5,
    "MASTERBUNDLE": 6,
    "TRACKGROUP": 7
}


# ----- Write modes -----
OVERWRITE = 'overwrite'
UPSERT = 'upsert'
APPEND = 'append'
ERROR = 'error'
IGNORE = 'ignore'

# ----- Options -----
OVERWRITE_SCHEMA = 'overwriteSchema'

# ----- formats -----
DELTA = "delta"
PARQUET = "parquet"




from pyspark.sql import DataFrame
from tidal_per_transformers.transformers import LoggableTransformer


class WithColumnRenamedTransformer(LoggableTransformer):
    """Transformer to rename a column

    """

    def __init__(self, source_column_name: str, target_column_name: str):
        super().__init__()
        self.source_column_name = source_column_name
        self.target_column_name = target_column_name

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return dataset.withColumnRenamed(self.source_column_name, self.target_column_name)



import pyspark.sql.types as T
import tidal_per_transformers.transformers.utils.constants as c


ARTISTS_SCHEMA = T.ArrayType(T.StructType([
    T.StructField(c.ALIASES, T.ArrayType(T.StringType())),
    T.StructField(c.ID, T.IntegerType()),
    T.StructField(c.NAME, T.StringType()),
    T.StructField(c.MAIN, T.BooleanType()),
    T.StructField(c.PICTURE, T.StringType())
]))


PLAYLIST_TRACKS_SCHEMA = T.ArrayType(T.StructType([
    T.StructField(c.INDEX, T.IntegerType()),
    T.StructField(c.PRODUCT_ID, T.IntegerType()),
    T.StructField(c.TITLE, T.StringType()),
    T.StructField(c.TRACK_GROUP, T.StringType()),
    T.StructField(c.ALBUM_ID, T.IntegerType()),
    T.StructField(c.MASTER_BUNDLE_ID, T.StringType()),
    T.StructField(c.ARTIST_ID, T.IntegerType()),
    T.StructField(c.DURATION, T.DoubleType()),
    T.StructField(c.AUDIO_QUALITY, T.StringType()),
    T.StructField(c.GENRE, T.StringType())]))



from typing import List

from pyspark.ml import Transformer


class SingleDAGPipeline:
    """This class creates a pipeline based on Transformers class
    The main issue with pyspark.ml.pipeline is that each stage creates its own DAG, which does not
    utilize spark new features like AQE and can lead to performance issues.
    """
    def __init__(self, stages: List[Transformer]):
        """
        :param stages: list of Transformers
        """
        self.stages = stages

    def fit(self, dataset):
        """
        :param dataset: DataFrame
        :return: PipelineModel
        """
        for stage in self.stages:
            dataset = stage.transform(dataset)
        return dataset



from .loggable_transformer import LoggableTransformer
from .aggregate_transformer import AggregateTransformer
from .artist_compound_transformer import ArtistCompoundMappingTransformer
from .artist_filter_transformer import ArtistFilterTransformer
from .artists_struct_filter_transformer import ArtistsStructFilterTransformer
from .blazingtext_input_format_transformer import BlazingTextInputFormatTransformer
from .clean_text_transformer import CleanTextTransformer
from .collect_ranked_list_transformer import CollectRankedListTransformer
from .deduplicate_sequence_transformer import DuplicateSequenceTransformer
from .distinct_sequence_transformer import DistinctSequenceTransformer
from .distinct_transformer import DistinctTransformer
from .drop_duplicates_transformer import DropDuplicatesTransformer
from .filter_infrequent_seq_transformer import FilterInfrequentSeqItemsTransformer
from .filter_infrequent_transformer import FilterInfrequentTransformer
from .flatten_struct_transformer import FlattenStructTransformer
from .join_transformer import JoinTransformer
from .minmax_scaling_transformer import MinMaxScalingTransformer
from .order_by_transformer import OrderByTransformer
from .participating_artist_transformer import ParticipatingArtistsTransformer
from .pivot_transformer import PivotTransformer
from .playlist_coherence_transformer import PlaylistCoherenceTransformer
from .replace_null_values_transformer import ReplaceNullValuesTransformer
from .select_transformer import SelectTransformer
from .sequence_content_filter_transformer import SequenceContentFilterTransformer
from .single_artist_filter_transformer import SingleArtistFilterTransformer
from .stack_transformer import StackTransformer
from .top_items_dithering_transformer import TopItemsDitheringTransformer
from .top_items_transformer import TopItemsTransformer
from .track_group_availability_transformer import TrackGroupAvailabilityTransformer
from .track_group_availability_by_country_transformer import TrackGroupAvailabilityByCountryTransformer
from .track_group_filter_transformer import TrackGroupFilterTransformer
from .union_transformer import UnionTransformer
from .voiceness_filter_transformer import VoicenessFilterTransformer
from .where_transformer import WhereTransformer
from .with_column_renamed_transfomer import WithColumnRenamedTransformer
from .with_column_transfomer import WithColumnTransformer
from .collect_ranked_map_transformer import CollectRankedMapTransformer
from .broadcast_array_intersect_transformer import BroadcastArrayIntersectTransformer
from .drop_columns_transformer import DropColumnsTransformer
from .filter_inactive_users_transformer import FilterInactiveUserTransformer



from pyspark.sql.window import Window
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer
import pyspark.sql.functions as F


class FilterInfrequentTransformer(LoggableTransformer):
    """
    Remove the items with a count smaller than cutoff

    :param columns: The columns used to group the dataset
    :param cutoff:  The minimum allowed group size
    """
    def __init__(self, columns, cutoff):
        super(FilterInfrequentTransformer, self).__init__()
        self.columns = columns
        self.cutoff = cutoff

    def _transform(self, dataset):
        window_spec = Window.partitionBy(self.columns)

        return (dataset
            .withColumn("cnt", F.count("*").over(window_spec))
            .where(F.col("cnt") >= self.cutoff)
            .drop("cnt"))



import pyspark.sql.functions as F
import tidal_per_transformers.transformers.utils.constants as c
from pyspark.sql import DataFrame

from tidal_per_transformers.transformers import LoggableTransformer


class BroadcastArrayIntersectTransformer(LoggableTransformer):
    def __init__(self, array_col: str, items_to_keep: DataFrame):
        """
        Transformer to filter out invalid items from an array column based on items_to_keep.
        :param array_col:       name of column containing an array of items
        :param items_to_keep:   a dataframe with only one row containing the list of items to be kept
        """
        super().__init__()
        self.array_col = array_col
        self.items_to_keep = items_to_keep.select(c.ITEMS_TO_KEEP)

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return (dataset
                .join(F.broadcast(self.items_to_keep))
                .withColumn(self.array_col, F.array_intersect(self.array_col, c.ITEMS_TO_KEEP))
                .drop(c.ITEMS_TO_KEEP))



from typing import List

from pyspark.ml.base import Transformer
from pyspark.sql import DataFrame


class DropColumnsTransformer(Transformer):
    def __init__(self, columns: List[str]):
        self.columns = columns
        assert isinstance(self.columns, List), "columns should be a list"
        super(DropColumnsTransformer, self).__init__()

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return dataset.drop(*self.columns)



from pyspark.sql import DataFrame

from tidal_per_transformers.transformers import LoggableTransformer


class DistinctTransformer(LoggableTransformer):
    """Transformer to get distinct rows
    """
    def __init__(self):
        super().__init__()

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return dataset.distinct()



from abc import abstractmethod

import pyspark.sql.functions as F
from pyspark.sql.dataframe import DataFrame

import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer
from tidal_per_transformers.transformers.track_group_filter_transformer import apply_category_filters


class ArtistFilterTransformer(LoggableTransformer):
    """
    Transformer for removing artists based on either category, popularity or a combination of the two. E.g. if you
    want the artist recommender feature to not recommend children music you can choose to filter all artists tagged
    with the children music tag. The popularity features are based on streaming data from the last 12 months
    """

    def __init__(self,
                 artist_filters: DataFrame,
                 remove_children_music: bool,
                 remove_ambient_music: bool,
                 remove_holiday_music: bool,
                 min_artist_streamers: int,
                 min_artist_streams: int):
        """
        :param artist_filters:          DataFrame containing the content filter information
        :param remove_ambient_music:    flag for toggling ambient music on/off
        :param remove_holiday_music:    flag for toggling holiday music on/off
        :param remove_children_music:   flag for toggling children music on/off
        :param min_artist_streams:      minimum number of streams for an artist to be included
        :param min_artist_streamers:    minimum number of unique listeners for an artist to be included
        """
        super(ArtistFilterTransformer, self).__init__()
        self.artist_filters = artist_filters
        self.remove_children_music = remove_children_music
        self.remove_ambient_music = remove_ambient_music
        self.remove_holiday_music = remove_holiday_music
        self.min_artist_streamers = min_artist_streamers
        self.min_artist_streams = min_artist_streams

    @abstractmethod
    def _transform(self, dataset):
        ...

    @staticmethod
    def apply_filters(category_filters: DataFrame,
                      stream_count: int,
                      streamers_count: int,
                      drop_holiday: bool,
                      drop_ambient: bool,
                      drop_children: bool) -> DataFrame:
        """Applies filters given in a data frame

        :param category_filters:    dataframe containing filters
        :param stream_count:        min stream count
        :param streamers_count:     min streamers count
        :param drop_holiday:        flag to drop holiday
        :param drop_ambient:        flag to drop ambient music
        :param drop_children:       flag to drop children
        :return: cleaned dataframe
        """
        return apply_category_filters(dataframe=category_filters
                                      .where(F.col(c.AVAILABLE))
                                      .where(F.col(c.NON_MUSIC) == 0)
                                      .where(F.col(c.STREAM_COUNT) >= stream_count)
                                      .where(F.col(c.STREAMERS_COUNT) >= streamers_count),
                                      drop_holiday=drop_holiday,
                                      drop_ambient=drop_ambient,
                                      drop_children=drop_children).select(c.ARTIST_ID)



from pyspark.ml import Transformer


# TODO: add unit tests

class FilterMatchingTextTransformer(Transformer):
    """
    Filter out when the text is matched in the column.
    """

    def __init__(self, column_name, texts, exact_match=False, case_sensitive=False):
        """
        :param column_name: name of the column to find the text
        :param text: list of text to find
        :param exact_match: should find exact text or surround with the wildcard %
        :param case_sensitive: match with the same case or independently
        """
        super(FilterMatchingTextTransformer, self).__init__()
        self.column_name = column_name
        self.texts = texts
        self.exact_match = exact_match
        self.case_sensitive = case_sensitive

    def _transform(self, dataset):

        if not self.case_sensitive:
            self.column_name = f"lower({self.column_name})"

        where = "1=1"

        for text in self.texts:

            if text is None:
                raise Exception("Cannot use NULL value in FilterMatchingTextTransformer")

            if not self.exact_match:
                text = f"%{text}%"

            if not self.case_sensitive:
                text = f"{text.lower()}"

            cond = f" and {self.column_name} not like '{text}'"

            where = where + cond

            # need to add not null or sql will remove all nulls
        where = f"({where}) or ({self.column_name} is null)"

        return dataset.where(where)



from typing import Optional, List, Union

from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer


class PivotTransformer(LoggableTransformer):

    def __init__(self,
                 group_by: Union[str, List],
                 pivot_col: str,
                 agg_exp,
                 values: Optional[List] = None):
        """
        Pivots a column of the current DataFrame and perform the specified aggregation

        :param group_by:    Columns to group the DataFrame by
        :param pivot_col:   Column used to pivot
        :param agg_exp:     The aggregation expression
        :param values:      A list of distinct values to pivot on
        """

        super().__init__()
        self.group_by = group_by
        self.pivot_col = pivot_col
        self.values = values
        self.agg_exp = agg_exp

    def _transform(self, dataset):
        return (dataset
                .groupBy(self.group_by)
                .pivot(self.pivot_col, self.values)
                .agg(self.agg_exp))



import pyspark.sql.functions as F
from pyspark.sql.dataframe import DataFrame

import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers.artist_filter_transformer import ArtistFilterTransformer
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer
from tidal_per_transformers.transformers.track_group_filter_transformer import TrackGroupFilterTransformer
from tidal_per_transformers.transformers.utils.schemas import PLAYLIST_TRACKS_SCHEMA

MATCHES = "matches"


class SequenceContentFilterTransformer(LoggableTransformer):
    """
    Transformer for applying content filters to columns with lists of tracks. This can e.g. be applied to the
    playlist track table containing a list of all the tracks in a playlist.
    """
    def __init__(self,
                 track_filters: DataFrame,
                 artist_filters: DataFrame,
                 remove_children_music: bool,
                 remove_ambient_music: bool,
                 remove_holiday_music: bool,
                 min_track_streams: int = 500,
                 min_track_streamers: int = 200,
                 min_track_duration: int = 45,
                 max_track_duration: int = 1200,
                 min_artist_streamers: int = 1000,
                 min_artist_streams: int = 2000,
                 schema=PLAYLIST_TRACKS_SCHEMA):
        """
        :param artist_filters:          DataFrame containing the content filter information
        :param remove_ambient_music:    flag for toggling ambient music on/off
        :param remove_holiday_music:    flag for toggling holiday music on/off
        :param remove_children_music:   flag for toggling children music on/off
        :param min_track_streams:       minimum number of recorded streams
        :param min_track_streamers:     minimum number of unique listeners
        :param min_track_duration:      minimum allowed duration of a track
        :param max_track_duration:      maximum allowed duration of a track
        :param min_artist_streams:      minimum number of streams for an artist to be included
        :param min_artist_streamers:    minimum number of unique listeners for an artist to be included
        """
        super(SequenceContentFilterTransformer, self).__init__()
        self.artist_filters = artist_filters
        self.track_filters = track_filters
        self.remove_children_music = remove_children_music
        self.remove_ambient_music = remove_ambient_music
        self.remove_holiday_music = remove_holiday_music
        self.min_artist_streamers = min_artist_streamers
        self.min_artist_streams = min_artist_streams
        self.min_track_streams = min_track_streams
        self.min_track_streamers = min_track_streamers
        self.min_track_duration = min_track_duration
        self.max_track_duration = max_track_duration
        self.schema = schema

    def _transform(self, dataset):
        artists_set = set(self.get_cleaned_artists().toPandas()[c.ARTIST_ID].tolist())
        artists_bc = dataset.sql_ctx.sparkSession.sparkContext.broadcast(artists_set)

        tracks_set = set(self.get_cleaned_track_groups().toPandas()[c.TRACK_GROUP].tolist())
        tracks_bc = dataset.sql_ctx.sparkSession.sparkContext.broadcast(tracks_set)

        @F.udf(returnType=self.schema)
        def filter_tracks(tracks):
            return [t for t in tracks if t[c.ARTIST_ID] in artists_bc.value and t[c.TRACK_GROUP] in tracks_bc.value]

        return (dataset
                .withColumn(c.TRACKS, filter_tracks(c.TRACKS))
                .where(F.size(c.TRACKS) > 0))  # Drop empty sequences

    def get_cleaned_artists(self):
        return ArtistFilterTransformer.apply_filters(self.artist_filters,
                                                     self.min_artist_streams,
                                                     self.min_artist_streamers,
                                                     self.remove_holiday_music,
                                                     self.remove_ambient_music,
                                                     self.remove_children_music).select(c.ARTIST_ID)

    def get_cleaned_track_groups(self):
        return TrackGroupFilterTransformer.apply_filters(self.track_filters,
                                                         self.min_track_streams,
                                                         self.min_track_streamers,
                                                         self.min_track_duration,
                                                         self.max_track_duration,
                                                         self.remove_holiday_music,
                                                         self.remove_ambient_music,
                                                         self.remove_children_music).select(c.TRACK_GROUP)



import pyspark.sql.functions as F
import tidal_per_transformers.transformers.utils.constants as c
from pyspark.ml.base import Transformer
from pyspark.sql import DataFrame

ARTIFACT_ID = "artifactId"
TRACK = "TRACK"
VIDEO = "VIDEO"
ARTIST = "ARTIST"
VIDEO_ID = "videoId"


class UserBlacklistFilterTransformer(Transformer):
    """ Transformer for removing content blacklisted by users from a DataFrame.
    """

    def __init__(self,
                 user_blacklist: DataFrame,  # backup_tables.get_user_blacklist_table(self.sc)
                 compound_table: DataFrame,  # backup_tables.get_artist_compound_mapping_table(self.sc)
                 filter_track: bool = True,
                 filter_artist: bool = True,
                 filter_video: bool = False,
                 user_col: str = c.USER_ID):
        super().__init__()
        self.filter_track = filter_track
        self.filter_video = filter_video
        self.filter_artist = filter_artist
        self.user_col = user_col
        self.user_blacklist = user_blacklist
        self.compound_table = compound_table

    def _transform(self, dataset):

        if self.user_col != c.USER_ID:
            self.user_blacklist = self.user_blacklist.withColumnRenamed(c.USER_ID, self.user_col)

        if self.filter_track:
            dataset = self.filter_blacklisted_tracks(dataset, self.user_blacklist)

        if self.filter_video:
            dataset = self.filter_blacklisted_videos(dataset, self.user_blacklist)

        if self.filter_artist:
            dataset = self.filter_blacklisted_artists(dataset, self.user_blacklist)

        return dataset

    def filter_blacklisted_tracks(self, dataset, blacklist):
        """
        Remove blacklisted tracks from the dataset. Maps the productIds -> trackGroups before removing the
        blacklisted content from the dataset.

        :type dataset:      pyspark.sql.DataFrame
        :param blacklist:   DataFrame containing all blacklisted content
        :type blacklist:    pyspark.sql.DataFrame
        """
        track_blacklist = (blacklist
                           .where(F.col(c.ARTIFACT_TYPE) == TRACK)
                           .select(self.user_col, F.col(ARTIFACT_ID).alias(c.TRACK_GROUP)))

        return dataset.join(track_blacklist, [self.user_col, c.TRACK_GROUP], "left_anti")

    def filter_blacklisted_videos(self, dataset, blacklist):
        """
        Remove blacklisted videos from the dataset.

        :type dataset:      pyspark.sql.DataFrame
        :param blacklist:   DataFrame containing all blacklisted content
        :type blacklist:    pyspark.sql.DataFrame
        """
        video_blacklist = (blacklist
                           .where(F.col(c.ARTIFACT_TYPE) == VIDEO)
                           .withColumnRenamed(ARTIFACT_ID, VIDEO_ID))

        return dataset.join(video_blacklist, [self.user_col, VIDEO_ID], "left_anti")

    def filter_blacklisted_artists(self, dataset, blacklist):
        """
        Remove blacklisted artists from the dataset. Filters any tracks by the artist or any compound the artist
        is part of.

        :type dataset:      pyspark.sql.DataFrame
        :param blacklist:   DataFrame containing all blacklisted content
        :type blacklist:    pyspark.sql.DataFrame
        """
        main_artist_blacklist = (blacklist
                                 .where(F.col(c.ARTIFACT_TYPE) == ARTIST)
                                 .select(F.col(self.user_col), F.col(ARTIFACT_ID).alias(c.ARTIST_ID))
                                 .persist())

        compound_table = (self.compound_table
                          .withColumnRenamed(c.ARTIST_ID, c.COMPOUND_ID))

        compound_blacklist = (main_artist_blacklist
                              .join(compound_table,
                                    main_artist_blacklist[c.ARTIST_ID] == compound_table[c.ARTIST_COMPOUND_ID])
                              .select(F.col(self.user_col), F.col(c.COMPOUND_ID).alias(c.ARTIST_ID)))

        # Keep both the main artists and any compounds they are part of
        full_blacklist = main_artist_blacklist.union(compound_blacklist)

        return dataset.join(full_blacklist, [self.user_col, c.ARTIST_ID], "left_anti")



import pyspark.sql.functions as F
from pyspark.sql.dataframe import DataFrame

import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer


class TrackGroupFilterTransformer(LoggableTransformer):
    """
    Transformer for removing track groups based on either category, popularity or a combination of the two. E.g. if you
    want to remove Christmas and children music from a track based model or mix you can use this transformer to easily
    exclude those content types. The popularity features are based on streaming data from the last 12 months
    """

    def __init__(self,
                 track_filters: DataFrame,
                 remove_children_music: bool,
                 remove_ambient_music: bool,
                 remove_holiday_music: bool,
                 min_track_streams: int = 500,
                 min_track_streamers: int = 200,
                 min_track_duration: int = 45,
                 max_track_duration: int = 1200):
        """
        :param track_filters:           DataFrame containing the content filter information
        :param remove_ambient_music:    flag for toggling ambient music on/off
        :param remove_holiday_music:    flag for toggling holiday music on/off
        :param remove_children_music:   flag for toggling children music on/off
        :param min_track_streams:       minimum number of recorded streams
        :param min_track_streamers:     minimum number of unique listeners
        :param min_track_duration:      minimum allowed duration of a track
        :param max_track_duration:      maximum allowed duration of a track
        """
        super(TrackGroupFilterTransformer, self).__init__()
        self.track_filters = track_filters
        self.remove_children_music = remove_children_music
        self.remove_ambient_music = remove_ambient_music
        self.remove_holiday_music = remove_holiday_music
        self.min_track_streams = min_track_streams
        self.min_track_streamers = min_track_streamers
        self.min_track_duration = min_track_duration
        self.max_track_duration = max_track_duration

    def _transform(self, dataset):
        assert c.TRACK_GROUP in dataset.columns
        return self.filter_tracks(dataset)

    def filter_tracks(self, tracks: DataFrame) -> DataFrame:
        cleaned_tracks = self.apply_filters(self.track_filters,
                                            self.min_track_streams,
                                            self.min_track_streamers,
                                            self.min_track_duration,
                                            self.max_track_duration,
                                            self.remove_holiday_music,
                                            self.remove_ambient_music,
                                            self.remove_children_music)

        return tracks.join(cleaned_tracks, c.TRACK_GROUP)

    @staticmethod
    def apply_filters(category_filters: DataFrame,
                      stream_count: int,
                      streamers_count: int,
                      min_track_duration: int,
                      max_track_duration: int,
                      drop_holiday: bool,
                      drop_ambient: bool,
                      drop_children: bool):
        return apply_category_filters(dataframe=category_filters
                                      .where(F.col(c.AVAILABLE))
                                      .where(F.col(c.NON_MUSIC) == 0)
                                      .where(F.col(c.STREAM_COUNT) >= stream_count)
                                      .where(F.col(c.STREAMERS_COUNT) >= streamers_count)
                                      .where(F.col(c.DURATION).between(min_track_duration, max_track_duration)),
                                      drop_holiday=drop_holiday,
                                      drop_ambient=drop_ambient,
                                      drop_children=drop_children).select(c.TRACK_GROUP)


def apply_category_filters(dataframe: DataFrame, drop_holiday: bool, drop_ambient: bool, drop_children: bool):
    if drop_children:
        dataframe = dataframe.where(F.col(c.CHILDREN) == 0)

    if drop_ambient:
        dataframe = dataframe.where(F.col(c.AMBIENT) == 0)

    if drop_holiday:
        dataframe = dataframe.where(F.col(c.HOLIDAY) == 0)

    return dataframe



from typing import Dict
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer


class WhereTransformer(LoggableTransformer):
    def __init__(self, where_filter):
        super().__init__()
        self.where_filter = where_filter

    def _transform(self, dataset):
        return dataset.where(self.where_filter)

    def get_loggable_dict(self, invocation_order: int) -> Dict:
        """ Returns a params dictionary that can be logged using TransformPipeLine class, with params items_column
            and field_column
        :param invocation_order: the invocation order of the transformer
        :return: loggable params dictionary
        """
        return {f'Transformer_{type(self).__name__}_{invocation_order}': {"where_clause": f'{self.where_filter}'}}



from pyspark.sql import functions as F
from pyspark.sql.types import MapType, ArrayType, StringType
from pyspark.sql.window import Window
from tidal_per_transformers.transformers import LoggableTransformer


class CollectRankedMapTransformer(LoggableTransformer):
    """
    Group a column by the 'group_by' column and collect the sorted values in the 'map_col' as a map

    :param group_by:        Columns to group the DataFrame by
    :param map_col:         Column for which we collect the values
    :param sort_col:        The column used to sort the values
    :param alias:           The name given to the new list column
    :param ascending:       Boolean to sort ascending ot descending
    :param limit:           Number of elements to keep in the list
    """

    def __init__(self,
                 group_by: any,
                 map_col: str,
                 sort_col: str,
                 ascending: bool = True,
                 alias: str = None,
                 limit=10_000):
        super().__init__()
        self.group_by = group_by
        self.sort_col = sort_col
        self.map_col = map_col
        self.alias = alias if alias else map_col
        self.ascending = ascending
        self.limit = limit

    def _transform(self, dataset):
        w = Window.partitionBy(self.group_by).orderBy(self.sort_col if self.ascending else F.desc(self.sort_col))

        return (dataset
                .withColumn(self.alias, F.collect_list(F.struct(self.map_col, self.sort_col)).over(w))
                .groupBy(self.group_by)
                .agg(F.max(self.alias).alias(self.alias))
                .withColumn(self.alias, F.slice(F.from_json(F.to_json(self.alias),
                                                            ArrayType(MapType(StringType(), StringType()))),
                                                1, self.limit))
                )



import unicodedata

import pandas as pd
import pyspark.sql.functions as F
from alphabet_detector import AlphabetDetector
from pyspark.ml.base import Transformer
from pyspark.sql.types import StringType

import tidal_per_transformers.transformers.utils.constants as c


@F.pandas_udf(returnType=StringType())
def strip_accents_udf(texts: pd.Series) -> pd.Series:

    def _strip_accents(text):
        if not AlphabetDetector().only_alphabet_chars(text, "LATIN"):
            return text

        try:
            # noinspection PyUnresolvedReferences
            text = unicode(text, 'utf-8')
        except NameError:  # unicode is a default on python 3
            pass

        return (unicodedata.normalize('NFD', text)
                .encode('ascii', 'ignore')
                .decode("utf-8"))

    return pd.Series([_strip_accents(text) for text in texts])


class CleanTextTransformer(Transformer):

    def __init__(self, text_col=c.TITLE, output_col=c.TITLE):
        super(CleanTextTransformer, self).__init__()
        self.text_col = text_col
        self.output_col = output_col

    def _transform(self, dataset):

        return (dataset
                .withColumn(self.output_col, F.regexp_replace(F.lower(F.col(self.text_col)), r'\([^)]*\)', ""))  # All text inside ()
                .withColumn(self.output_col, F.regexp_replace(F.col(self.output_col), r'\[[^\]]*\]', ""))  # All text inside []
                .withColumn(self.output_col, F.regexp_replace(F.col(self.output_col), r'/[^/]*$', ""))  # All text after /
                .withColumn(self.output_col, strip_accents_udf(self.output_col))
                .withColumn(self.output_col, F.trim(F.regexp_replace(F.col(self.output_col), r'[\?,.!\s]', ""))))  # Remove ?






from typing import List

from pyspark.sql import DataFrame
from tidal_per_transformers.transformers import LoggableTransformer


class DropDuplicatesTransformer(LoggableTransformer):
    """Transformer to drop duplicates

    :param columns: Columns to drop duplicates on
    """

    def __init__(self, columns: List[str] = None):
        super().__init__()
        self.columns = columns

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return dataset.dropDuplicates(self.columns)



from pyspark.sql.dataframe import DataFrame

import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers import ArtistFilterTransformer


class SingleArtistFilterTransformer(ArtistFilterTransformer):
    """
        :param artist_filters:          DataFrame containing the content filter information
        :param remove_ambient_music:    flag for toggling ambient music on/off
        :param remove_holiday_music:    flag for toggling holiday music on/off
        :param remove_children_music:   flag for toggling children music on/off
        :param min_artist_streams:      minimum number of streams for an artist to be included
        :param min_artist_streamers:    minimum number of unique listeners for an artist to be included
        :param artist_column:           Column containing artists
        """

    def __init__(self,
                 artist_filters: DataFrame,
                 remove_children_music: bool,
                 remove_ambient_music: bool,
                 remove_holiday_music: bool,
                 min_artist_streamers: int = 500,
                 min_artist_streams: int = 2000,
                 artist_column: str = c.ARTIST_ID
                 ):
        super().__init__(artist_filters,
                         remove_children_music,
                         remove_ambient_music,
                         remove_holiday_music,
                         min_artist_streamers,
                         min_artist_streams)
        self.artist_filters = artist_filters
        self.remove_children_music = remove_children_music
        self.remove_ambient_music = remove_ambient_music
        self.remove_holiday_music = remove_holiday_music
        self.min_artist_streamers = min_artist_streamers
        self.min_artist_streams = min_artist_streams
        self.artist_column = artist_column

    def _transform(self, dataset):
        cleaned_artists = self.apply_filters(self.artist_filters,
                                             self.min_artist_streams,
                                             self.min_artist_streamers,
                                             self.remove_holiday_music,
                                             self.remove_ambient_music,
                                             self.remove_children_music)
        return (dataset
                .join(cleaned_artists.withColumnRenamed(c.ARTIST_ID, self.artist_column),
                      self.artist_column,
                      how="inner"))



from typing import Optional, List

import pyspark.sql.functions as F
from pyspark.sql.window import Window
from pyspark.sql import DataFrame

import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers import LoggableTransformer


class ArtifactIdMappingByUserCountryTransformer(LoggableTransformer):
    """
    The transformer maps trackGroup / masterAlbumId to the minimum productId / albumId
    that is available in the country of given user
    """
    def __init__(self,
                 item_metadata: DataFrame,
                 user_countries: DataFrame,
                 available_countries_col: str = c.AVAILABLE_COUNTRY_CODES,
                 user_country_col: str = c.COUNTRY_CODE,
                 id_col: str = c.ID,
                 item_id_col: str = c.TRACK_GROUP,
                 item_instance_id_col: str = c.PRODUCT_ID,
                 partition_cols: Optional[List[str]] = None):

        super().__init__()
        self.item_metadata = item_metadata.select(item_id_col,
                                                  F.col(id_col).alias(item_instance_id_col),
                                                  available_countries_col)
        self.user_countries = user_countries.select(c.USER_ID, user_country_col)
        self.item_id_col = item_id_col
        self.item_instance_id_col = item_instance_id_col
        self.available_countries_col = available_countries_col
        self.user_country_col = user_country_col
        self.partition_cols = partition_cols if partition_cols else [item_id_col, c.USER_ID]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        window = Window.partitionBy(self.partition_cols).orderBy(self.item_instance_id_col)
        return (dataset
                .join(self.user_countries, c.USER_ID)
                .join(self.item_metadata, self.item_id_col)
                .where(F.array_contains(self.available_countries_col, F.col(self.user_country_col)))
                .withColumn(c.RN, F.rank().over(window))
                .where(F.col(c.RN) == 1)
                .drop(self.user_country_col, self.available_countries_col, c.RN))



import pyspark.sql.functions as F
import tidal_per_transformers.transformers.utils.constants as c

from pyspark.sql import DataFrame
from tidal_per_transformers.transformers import LoggableTransformer


class TrackGroupAvailabilityByCountryTransformer(LoggableTransformer):

    def __init__(self,
                 track_group_availability: DataFrame,
                 country_col: str = c.COUNTRY_CODE,
                 availability_column: str = c.AVAILABLE_COUNTRY_CODES):
        """
        Filter any track group from the input dataset that is not available for streaming in the given country.
        This is used in personalized models to avoid recommending tracks that are not available to a user.

        :param track_group_availability:  all available track groups (source: track group availability job)
        :param country_col:               column containing the country code of the user
        :param availability_column:       column containing the country code the track is available in
        """
        super(TrackGroupAvailabilityByCountryTransformer, self).__init__()
        self.track_group_availability = track_group_availability
        self.country_col = country_col
        self.availability_column = availability_column

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return (dataset
                .join(self.track_group_availability, c.TRACK_GROUP)
                .where(F.array_contains(self.availability_column, F.col(self.country_col)))
                .drop(self.availability_column))



from pyspark.ml.base import Transformer


class StackTransformer(Transformer):

    def __init__(self,
                 cols_to_fix: list,
                 cols_to_stack: list,
                 output_col_names: list):
        """ Reshape dataframe into long format

        :param cols_to_fix:         columns that should remain
        :param cols_to_stack:       columns to stack. Column values will be stacked into values and column neames will
        be stacked into 'variable' column
        :param output_col_names:    names for value columns, typically [c.VARIABLE, c.VALUE]
        :param value_col_str:       string with names for columns that are stacked
        :param stack_expr:          string with stack expression, e.g.
        "stack(2, 'nStreams', nStreams, 'streamDuration', streamDuration) as (variable, value)"

        Before stacking

        | date       | experimentVariant | nUsers | streamDuration | sessionLength |
        |------------|-------------------|--------|----------------|---------------|
        | 2022-01-01 | control           | 1000   | 6000           | 30000         |
        | 2022-01-01 | test_variant1     | 2000   | 5000           | 49999         |
        | 2022-01-01 | test_variant2     | 2500   | 5000           | 49999         |

        After stacking

        | date       | experimentVariant | nUsers | variable       | value |
        |------------|-------------------|--------|----------------|-------|
        | 2022-01-01 | control           | 1000   | streamDuration | 6000  |
        | 2022-01-01 | test_variant1     | 2000   | streamDuration | 5000  |
        | 2022-01-01 | test_variant2     | 2500   | streamDuration | 5000  |
        | 2022-01-01 | control           | 1000   | sessionLength  | 30000 |
        | 2022-01-01 | test_variant1     | 2000   | sessionLength  | 49999 |
        | 2022-01-01 | test_variant2     | 2500   | sessionLength  | 49999 |

        cols_to_fix = ["date", "experimentVariant", "nUsers"]
        cols_to_stack = ["streamDuration", "sessionLength"]
        output_col_names = ["variable", "value"]
        """
        super().__init__()
        self.cols_to_stack = cols_to_stack
        self.cols_to_fix = cols_to_fix
        self.output_col_names = output_col_names
        assert len(self.output_col_names) == 2, "Should not pass more than 2 values here!"

    def _transform(self, dataset):
        return dataset.selectExpr(*self.cols_to_fix, self._get_stack_expr())

    def _get_stack_expr(self):
        value_col_str = "(" + ','.join(self.output_col_names) + ")"
        return (f'stack({len(self.cols_to_stack)},' + ','.join(f'"{i}", {i}' for i in self.cols_to_stack) + ') as '
                + value_col_str)



from typing import Any

from pyspark.sql import DataFrame
from tidal_per_transformers.transformers import LoggableTransformer


class WithColumnTransformer(LoggableTransformer):
    """Create a new column with a given pyspark expression.

    """

    def __init__(self, column_name: str, expression: Any):
        super().__init__()
        self.column_name = column_name
        self.expression = expression

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return dataset.withColumn(self.column_name, self.expression)



from datetime import datetime, timedelta

import pyspark.sql.functions as F
from dateutil.relativedelta import relativedelta
import tidal_per_transformers.transformers.utils.constants as c

from tidal_per_transformers.transformers import LoggableTransformer

DEATH_DATE_BUFFER_DAYS = 180  # Time window before dead artists are removed
ACTIVE_YEARS_BUFFER = 1

METADATA_DEATH_DATE = "Death Date"
METADATA_ACTIVE_YEARS = "Active Years"


class FilterDeadAndInactiveArtistsTransformer(LoggableTransformer):
    """
    Filter out releases from artists which are either dead or no longer active
    """
    def __init__(self, rovi_metadata, artist_column=c.ARTIST_ID):
        super().__init__()
        self.artist_column = artist_column
        self.rovi_metadata = rovi_metadata

    # TODO: use artists feature store
    def _transform(self, dataset):
        metadata = (self.rovi_metadata
                    .where(F.col(c.TYPE).isin([METADATA_DEATH_DATE, METADATA_ACTIVE_YEARS]))
                    .withColumnRenamed(c.ARTIFACT_ID, self.artist_column)
                    .withColumnRenamed(c.TYPE, c.METADATA_TYPE)
                    .withColumnRenamed(c.VALUE, c.METADATA_VALUE))

        living_artists = self._filter_dead_artists(dataset, metadata, DEATH_DATE_BUFFER_DAYS)
        return self._filter_non_active_artists(living_artists, metadata, ACTIVE_YEARS_BUFFER)

    def _filter_dead_artists(self, dataset, metadata, buffer_days):
        cut_off = (datetime.today() - timedelta(days=buffer_days)).date()

        death_dates = metadata.where(F.col(c.METADATA_TYPE) == METADATA_DEATH_DATE)

        dead_artists = (death_dates
                        .filter(f"'{cut_off}' > {c.METADATA_VALUE}")
                        .select(F.col(self.artist_column)))

        return dataset.join(dead_artists, self.artist_column, "left_anti")

    def _filter_non_active_artists(self, dataset, metadata, delay_years):
        cut_off = (datetime.today() - relativedelta(years=delay_years)).date().year
        active_years = metadata.where(F.col(c.METADATA_TYPE) == METADATA_ACTIVE_YEARS)
        valid_entries = active_years.where(F.length(F.col(c.METADATA_VALUE)) == 21)

        last_active_year = (valid_entries
                            .withColumn("lastActiveYear", F.substring(F.split(c.METADATA_VALUE, "_")[1], 0, 4))
                            .groupBy(self.artist_column).agg(F.max("lastActiveYear").alias("max"))
                            .filter(F.col("max") <= cut_off)
                            .select(F.col(self.artist_column)))

        return dataset.join(last_active_year, self.artist_column, "left_anti")



from abc import ABC
from typing import Dict
from pyspark.ml.base import Transformer
import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers.utils.class_utils import as_dict


class LoggableTransformer(Transformer, ABC):
    """
    Collect list on "item_column" and ordered by order_by_column
    """
    def __init__(self):
        super(LoggableTransformer, self).__init__()

    def get_loggable_dict(self, invocation_order: int) -> Dict:
        """
        Collects the loggable dict for the given transformer with its invocation order.
        :param invocation_order: the invocation order of the given transformer
        :return: loggable params dictionary
        """
        params = as_dict(self)
        [params.pop(x, None) for x in [c.UID, c.HIDDEN_PARAMS]]
        return {f'Transformer_{type(self).__name__}_{invocation_order}': params}



import pyspark.sql.functions as F
import tidal_per_transformers.transformers.utils.constants as c

from pyspark.sql import DataFrame
from tidal_per_transformers.transformers import LoggableTransformer


class ParticipatingArtistsTransformer(LoggableTransformer):
    """Transforms artists field into main and featuring artist transformer

    """

    def __init__(self, artist_field: str = 'artists'):
        super().__init__()
        self.main_artist_filter = f"filter({artist_field}, x -> x.main == 'true')"
        self.featuring_artist_filter = f"filter({artist_field}, x -> x.main == 'false')"

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return (dataset
                .withColumn(c.MAIN_ARTISTS, F.expr(self.main_artist_filter))
                .withColumn(c.FEATURING_ARTISTS, F.expr(self.featuring_artist_filter))
                )



from itertools import chain

import pyspark.sql.functions as F
from pyspark.sql import Column
from pyspark.sql import Window
from pyspark.sql.types import MapType, StringType


def get_top_items(df, partition, order, n):
    """
    Get the top 'n' items per partitions from the given ordering

    :param df:          DataFrame to search
    :param partition:   Column(s) to partition by
    :param order:       Column(s) to order by
    :param n:           Number of top items to return per partition
    :return:            DataFrame containing the top n items per partition
    """
    window = Window.partitionBy(partition).orderBy(order)
    return df.withColumn("rn", F.row_number().over(window)).where(F.col("rn") <= n).drop("rn")


def list_to_index_map(l):
    return {str(i): l[i] for i in range(0, len(l))}


list_to_top_list = F.udf(list_to_index_map, MapType(StringType(), StringType()))


def dict_to_map_column(mapping: dict) -> Column:
    """
    Function to create map from dictionary, used e.g. to create a column from a dictionary.

    :param mapping: dictionary of key value pairs to be used to create new column.
    :return column
    """
    return F.create_map([F.lit(x) for x in chain(*mapping.items())])



from pyspark.sql import DataFrame
from pyspark.sql.window import Window
import pyspark.sql.functions as F
import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers import LoggableTransformer


class MultipleCompoundTransformer(LoggableTransformer):
    """
    Custom compound mapping transformer mapping compoundIds to all main artists on a track
    For tracks with > 1 main artists they are all assigned a weight equal to count/num_artists
    E.g. for each stream on a track featuring both Jay-Z and Kanye they each receive 0.5 streams each
    """
    def __init__(self, compound_mapping: DataFrame):
        super().__init__()
        self.compound_mapping = compound_mapping

    def _transform(self, dataset) -> DataFrame:
        main_artist_mapping = self.get_main_artist_counts()

        return (dataset
                .join(main_artist_mapping, c.ARTIST_ID, "left")
                .withColumn(c.COUNT, F.when(F.col(c.ARTIST_COMPOUND_ID).isNull(),
                                            F.col(c.COUNT)).otherwise(F.col(c.COUNT) / F.col(c.NUM_MAIN_ARTISTS)))
                .withColumn(c.ARTIST_ID, F.when(F.col(c.ARTIST_COMPOUND_ID).isNull(),
                                                F.col(c.ARTIST_ID)).otherwise(F.col(c.ARTIST_COMPOUND_ID)))
                .drop(c.ARTIST_COMPOUND_ID, c.NUM_MAIN_ARTISTS))

    def get_main_artist_counts(self) -> DataFrame:
        artist_window = Window.partitionBy(c.ARTIST_ID)

        return (self.compound_mapping
                .where(F.col(c.MAIN_ARTIST))
                .select(c.ARTIST_ID, c.ARTIST_COMPOUND_ID)
                .withColumn(c.NUM_MAIN_ARTISTS, F.count("*").over(artist_window)))


from datetime import date


def as_dict(obj):
    # TODO remove from ml delta table
    """ Convert the class object to a dictionary with all members and it's values """
    d = {}
    for k, v in obj.__dict__.items():
        if isinstance(v, dict):
            for h, val in v.items():
                d[f"{k}_{h}"] = val
        elif isinstance(v, (bool, list, date)):
            d[k] = str(v)
        elif isinstance(v, (str, float, int)):
            d[k] = v
        elif isinstance(v, type(None)):
            d[k] = "None"
    return d



from typing import List
import tidal_per_transformers.transformers.utils.constants as c
from pyspark.ml.base import Transformer
from pyspark.sql import functions as F
from pyspark.sql.window import Window


class DiversitySortTransformer(Transformer):
    """
    Sorts a mix in an 'intelligent' manner by e.g. spacing out tracks from the same artists or whatever you
    pass using the partition arguments.

    :param id_col:          Your id, this could e.g. be a mixId or userId or another partition we want to sort for
    :param partition_one:   The first partition you would like space out (e.g. artistId)
    :param partition_two:   The second partition you would like to space out (e.g. trackBundleId)
    :param order_by:        Column containing the original relevance based ordering
    :param gap:             The number of spaces to add between each artist/album recommendation
    :return:                A DataFrame with a new 'rank' column that can be used to sort the list
    """

    def __init__(self,
                 id_col: List[str],
                 partition_one: List[str],
                 partition_two: List[str], order_by, gap=5):
        super(DiversitySortTransformer, self).__init__()
        self.id_col = id_col
        self.partition_one = self.id_col + partition_one
        self.partition_two = self.id_col + partition_two
        self.order_by = order_by
        self.gap = gap

    def _transform(self, dataset):
        w1 = Window.partitionBy(self.partition_one).orderBy(self.order_by)
        w1_rank_window = Window.partitionBy(self.id_col).orderBy("w1_first_rank")
        w2 = Window.partitionBy(self.partition_two).orderBy(self.order_by)
        w2_rank_window = Window.partitionBy(self.id_col).orderBy("w2_first_rank")

        ordered = (dataset
                   .withColumn("w1_first_rank", F.first(self.order_by, True).over(w1) - 1)
                   .withColumn("w2_first_rank", F.first(self.order_by, True).over(w2) - 1)
                   .withColumn("w1_rank", F.dense_rank().over(w1_rank_window) - 1)
                   .withColumn("w2_rank", F.dense_rank().over(w2_rank_window) - 1)
                   .withColumn("w1_inter_rank", F.row_number().over(w1) - 1)
                   .withColumn("w2_inter_rank", F.row_number().over(w2) - 1)
                   .withColumn("ordering", (F.least(F.col("w1_rank"), F.col("w2_rank"))
                                            + self.gap * F.greatest(F.col("w1_inter_rank"), F.col("w2_inter_rank"))))
                   .withColumn(c.RANK, F.row_number().over(Window.partitionBy(self.id_col).orderBy(
                    F.col('ordering'), F.desc(self.order_by))))
                   .drop("w1_inter_rank", "w2_inter_rank", "w2_rank", "w1_rank", "w1_first_rank", "w2_first_rank",
                         "ordering"))
        return ordered



from pyspark.ml.base import Transformer
from pyspark.sql import functions as F, DataFrame

import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers.utils import get_top_items


class MainArtistCompoundMappingTransformer(Transformer):
    """
    Map the artist compound id's to a single main artist (e.g. Miguel feat. Travis Scott -> Miguel)

    :returns    DataFrame where the compound id's have been mapped to their constituent parts
    """
    def __init__(self, artist_compound_mapping: DataFrame):
        super().__init__()
        self.artist_compound_mapping = artist_compound_mapping

    def _transform(self, dataset):
        compound_map = (self.artist_compound_mapping
                        .where("mainartist = 'true'")
                        .withColumnRenamed(c.ARTIST_COMPOUND_ID, c.RESOLVED_ARTIST_ID)
                        .drop(c.ID))

        # Unfortunately the compound table contain duplicates (multiple main artists), keep only 1 (lowest priority)
        deduped = get_top_items(compound_map, [c.ARTIST_ID], c.PRIORITY, 1)

        # If there is no compound entry we already have the main artist
        joined = dataset.join(deduped, c.ARTIST_ID, "left")

        mapped = (joined
            .withColumn(c.ARTIST_ID, F.when(
                F.col(c.RESOLVED_ARTIST_ID).isNull(), F.col(c.ARTIST_ID))
                    .otherwise(F.col(c.RESOLVED_ARTIST_ID)))
            .drop(c.RESOLVED_ARTIST_ID, c.PRIORITY, c.MAIN_ARTIST, c.ARTIST_COMPOUND_ID))

        return mapped



from typing import Optional, List

import pyspark.sql.functions as F
from pyspark.sql.dataframe import DataFrame

import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer
from tidal_per_transformers.transformers.track_group_filter_transformer import apply_category_filters


class MasterBundleFilterTransformer(LoggableTransformer):
    """
    Transformer for removing albums based on either category,
    popularity (i.e., # streams and # streamers from the last 12 months) or a combination of the two.
    """

    def __init__(self,
                 album_filters: DataFrame,
                 remove_children_music: bool,
                 remove_ambient_music: bool,
                 remove_holiday_music: bool,
                 min_album_streamers: int,
                 min_album_streams: int,
                 album_column: str = c.MASTER_BUNDLE_ID,
                 album_types: Optional[List[str]] = None):
        """
        :param album_filters:          DataFrame containing the content filter information
        :param remove_ambient_music:    flag for toggling ambient music on/off
        :param remove_holiday_music:    flag for toggling holiday music on/off
        :param remove_children_music:   flag for toggling children music on/off
        :param min_album_streamers:     minimum number of streams for an album to be included
        :param min_album_streams:       minimum number of unique listeners for an album to be included
        :param album_column:            Column containing master bundle id
        """
        super().__init__()
        self.album_filters = album_filters
        self.remove_children_music = remove_children_music
        self.remove_ambient_music = remove_ambient_music
        self.remove_holiday_music = remove_holiday_music
        self.min_album_streamers = min_album_streamers
        self.min_album_streams = min_album_streams
        self.album_column = album_column
        self.album_types = album_types or [c.ALBUM.upper()]

    def _transform(self, dataset):
        cleaned_albums = self.apply_filters(self.album_filters,
                                            stream_count=self.min_album_streams,
                                            streamers_count=self.min_album_streamers,
                                            album_types=self.album_types,
                                            drop_holiday=self.remove_holiday_music,
                                            drop_ambient=self.remove_ambient_music,
                                            drop_children=self.remove_children_music)

        return (dataset
                .join(cleaned_albums.withColumnRenamed(c.MASTER_BUNDLE_ID, self.album_column),
                      self.album_column))

    @staticmethod
    def apply_filters(category_filters: DataFrame,
                      stream_count: int,
                      streamers_count: int,
                      album_types: List[str],
                      drop_holiday: bool,
                      drop_ambient: bool,
                      drop_children: bool) -> DataFrame:
        """Applies filters given in a data frame

        :param category_filters:    dataframe containing filters
        :param stream_count:        min stream count
        :param streamers_count:     min streamers count
        :param album_types:         list of album types
        :param drop_holiday:        flag to drop holiday
        :param drop_ambient:        flag to drop ambient music
        :param drop_children:       flag to drop children
        :return: cleaned dataframe
        """
        return apply_category_filters(dataframe=category_filters
                                      .where(F.col(c.AVAILABLE))
                                      .where(F.col(c.ALBUM_TYPE).isin(album_types))
                                      .where(F.col(c.NON_MUSIC) == 0)
                                      .where(F.col(c.STREAM_COUNT) >= stream_count)
                                      .where(F.col(c.STREAMERS_COUNT) >= streamers_count),
                                      drop_holiday=drop_holiday,
                                      drop_ambient=drop_ambient,
                                      drop_children=drop_children).select(c.MASTER_BUNDLE_ID)



from tidal_per_transformers.transformers import LoggableTransformer


class AggregateTransformer(LoggableTransformer):
    """
    Group a column by the 'group_by' column and perform aggregation expressions defined in 'agg_exp'.
    This implementation does unfortunately not support multiple aggregations per column!

    :param group_by:    Columns to group the DataFrame by
    :type group_by:     str|list
    :param agg_exp:     The aggregation expression
    :type agg_exp:      dict|list
    """
    def __init__(self, group_by, agg_exp, aliases=None):
        super(AggregateTransformer, self).__init__()
        self.group_by = group_by
        self.agg_exp = agg_exp
        self.aliases = aliases
        if aliases is not None:
            assert len(aliases) == len(agg_exp), "Please provide an alias for every aggregated column"

    def _transform(self, dataset):
        if type(self.agg_exp) is list:
            output = dataset.groupBy(self.group_by).agg(*self.agg_exp)
        else:
            output = dataset.groupBy(self.group_by).agg(self.agg_exp)

        return output if self.aliases is None else self.rename_agg_columns(output)

    def rename_agg_columns(self, dataset):
        columns = {f'{v}({k})': alias for alias, (k, v) in zip(self.aliases, self.agg_exp.items())}
        # rename won't work for count(*) - correct for this
        if 'count(*)' in columns:
            columns['count(1)'] = columns.pop('count(*)')

        for old_name, new_name in columns.items():
            dataset = dataset.withColumnRenamed(old_name, new_name)
        return dataset





from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer


class SelectTransformer(LoggableTransformer):
    def __init__(self, selected_columns):
        super().__init__()
        self.selected_columns = selected_columns

    def _transform(self, dataset):
        return dataset.select(self.selected_columns)



import math

import numpy as np
import pyspark.sql.functions as F
from pyspark.sql import Window
from pyspark.sql.types import DoubleType


def dither(dataset, rank_col, partition_by, epsilon=2.0, start_pos=0, dither_col=None):
    """
    Dither the 'rank_col' in the DataFrame and returns the dithering ordering.

    :param dataset:         The dataset containing the original ordering
    :type dataset:          pyspark.sql.DataFrame
    :param rank_col:        The original rank
    :param partition_by:    The partition we want to dither
    :type partition_by:     list(str)
    :param epsilon:         Controls the amount of dithering applied (range 1.25 - 2.0)
    :param start_pos:       Controls the amount of dithering applied to the top items (higher = more)
    :param dither_col:      Column name to put the dithered number
    :rtype:                 pyspark.sql.DataFrame
    """
    sd = dither_sd(epsilon)

    if dither_col is None:
        dither_col = rank_col

    return (dataset
            .withColumn("sd", F.lit(sd))
            .withColumn("dither_score", dither_score_udf(rank_col, "sd", F.lit(start_pos)))
            .withColumn(dither_col, F.row_number().over(Window.partitionBy(partition_by).orderBy("dither_score")))
            .drop("sd", "dither_score"))


def dither_sd(epsilon):
    return math.sqrt(np.log(epsilon)) if epsilon > 1.0 else 1e-10


def dither_score(pos, sd, x_start=0):
    """ Generate a dithering score that determines the new order of the rows.
    Source: https://buildingrecommenders.wordpress.com/2015/11/11/dithering/

    :param pos:       The original position in the mix (starting from 1)
    :param sd:        The standard deviation of the distribution
    :param x_start:   The starting position on the x-axis
    """
    return float((np.log(x_start + pos) + np.random.normal(0.0, sd, 1))[0])


dither_score_udf = F.udf(dither_score, DoubleType())



from pyspark.sql import DataFrame

from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer


class OrderByTransformer(LoggableTransformer):
    def __init__(self, subset):
        super().__init__()
        self.subset = subset

    def _transform(self, dataset: DataFrame) -> DataFrame:
        return dataset.orderBy(self.subset)



from typing import Optional

import pyspark.sql.functions as F
from tidal_per_transformers.transformers.loggable_transformer import LoggableTransformer
from pyspark.sql.types import DataType

import tidal_per_transformers.transformers.utils.constants as c


class FilterInfrequentSeqItemsTransformer(LoggableTransformer):

    def __init__(self,
                 min_item_frequency: int,
                 sequence_column: str,
                 item_column: Optional[str],
                 sequence_id_column: str,
                 sequence_schema: DataType):
        """
        Filter out less frequent tracks from each playlist

        :param min_item_frequency:      Minimum number of sequences containing the item
        :param sequence_column:         Column name of the sequence, e.g., "tracks"
        :param item_column:             Column name of item field in the sequence (assume tracks are structs)
        :param sequence_id_column:      Column name of playlist id
        :param sequence_schema:         Data type of tracks column
        """
        super().__init__()
        self.min_item_frequency = min_item_frequency
        self.sequence_column = sequence_column
        self.item_column = item_column
        self.sequence_id_column = sequence_id_column
        self.sequence_schema = sequence_schema

    def _transform(self, dataset):
        explode_col = f"{self.sequence_column}.{self.item_column}" if self.item_column else self.sequence_column

        items = (dataset
                 .select(F.explode(F.col(explode_col)).alias(c.ITEM), F.col(self.sequence_id_column))
                 .groupBy(c.ITEM)
                 .agg(F.count(self.sequence_id_column).alias(c.COUNT))
                 .where(F.col(c.COUNT) >= self.min_item_frequency)
                 .select(c.ITEM)).toPandas()[c.ITEM].tolist()

        items_bc = dataset.sql_ctx.sparkSession.sparkContext.broadcast(set(items))

        @F.udf(returnType=self.sequence_schema)
        def filter_tracks(sequence):
            if self.item_column:
                return [item for item in sequence if item[self.item_column] in items_bc.value]
            else:
                return [item for item in sequence if item in items_bc.value]

        return dataset.withColumn(self.sequence_column, filter_tracks(self.sequence_column))



from pyspark.sql import functions as F
from pyspark.sql.window import Window
from tidal_per_transformers.transformers import LoggableTransformer


class CollectRankedListTransformer(LoggableTransformer):
    """
    Group a column by the 'group_by' column and collect the sorted values in the 'list_col' as a list

    :param group_by:   Columns to group the DataFrame by
    :param list_col:        Column for which we collect the values
    :param sort_by:         The column used to sort the values
    :param alias:           The name given to the new list column
    """
    def __init__(self, group_by, list_col, sort_by, alias):
        super(CollectRankedListTransformer, self).__init__()
        self.group_by = group_by
        self.list_col = list_col
        self.sort_by = sort_by
        self.alias = alias
        assert(isinstance(group_by, list))

    def _transform(self, dataset):
        w = Window.partitionBy(self.group_by).orderBy(self.sort_by)

        return (dataset
                .withColumn(self.alias, F.collect_list(self.list_col).over(w))
                .groupBy(self.group_by)
                .agg(F.max(self.alias).alias(self.alias)))



from typing import Optional

import pyspark.sql.functions as F
from pandas import DataFrame
from pyspark.ml import Transformer

import tidal_per_transformers.transformers.utils.constants as c


class VoicenessFilterTransformer(Transformer):

    def __init__(self, voiceness_scores: DataFrame, track_index: DataFrame, score_threshold: Optional[float] = None):
        """
        Transformer used to remove tracks containing voice (non music) content

        :parma voice_scores:     dataframe containing the predicted voiceness scores
        :param track_index:      track index used to map productIds to trackGroup
        :param score_threshold:  custom threshold value used to determine classification (default = 0.5)
        """
        super().__init__()
        self.voiceness_scores = voiceness_scores
        self.product_to_track_group_mapping = track_index.select(F.col(c.ID).alias(c.PRODUCT_ID), c.TRACK_GROUP)
        self.score_threshold = score_threshold

    def _transform(self, dataset: DataFrame) -> DataFrame:
        assert c.TRACK_GROUP in dataset.columns, f"DataFrame must contain a {c.TRACK_GROUP} column"
        return dataset.join(self.get_voice_track_groups(self.voiceness_scores), c.TRACK_GROUP, "left_anti")

    def get_voice_track_groups(self, voiceness: DataFrame) -> DataFrame:
        if self.score_threshold:
            voiceness = voiceness.where(F.col(c.SCORE) >= self.score_threshold)
        else:
            voiceness = voiceness.where(F.col(c.VOICE) == 1)

        return (voiceness
                .join(self.product_to_track_group_mapping, c.PRODUCT_ID)
                .select(c.TRACK_GROUP)
                .distinct())




from typing import Dict
import pyspark.sql.functions as F

from pyspark.sql import DataFrame
from tidal_per_transformers.transformers import LoggableTransformer


class FlattenStructTransformer(LoggableTransformer):
    """Flatten each struct field into a new column

    :param struct_fields: Dictionary containing new column names and their corresponding struct field names
    """
    def __init__(self, struct_fields: Dict[str, str]):
        super().__init__()
        self.struct_fields = struct_fields

    def _transform(self, dataset: DataFrame) -> DataFrame:
        for key, value in self.struct_fields.items():
            dataset = dataset.withColumn(key, F.col(value))
        return dataset



from pyspark.ml import Transformer
from pyspark.sql.window import Window
from datetime import datetime, timedelta
import pyspark.sql.functions as F
import tidal_per_transformers.transformers.utils.constants as c


class FilterInactiveUserTransformer(Transformer):

    def __init__(self, max_inactivity_months):
        super(FilterInactiveUserTransformer, self).__init__()
        self.max_inactivity_months = max_inactivity_months

    def _transform(self, dataset):
        if self.max_inactivity_months is None or self.max_inactivity_months == 0:
            return dataset

        window_spec = Window.partitionBy(c.USER_ID)

        return (dataset
            .withColumn("lastActiveDate", F.max(c.LAST_STREAMED_DATE).over(window_spec))
            .where(F.col("lastActiveDate") >= datetime.now() - timedelta(days=self.max_inactivity_months*30))
            .drop("lastActiveDate"))


import pyspark.sql.functions as F
from pyspark.sql import DataFrame

import tidal_per_transformers.transformers.utils.constants as c
from tidal_per_transformers.transformers import LoggableTransformer


class TrackGroupAvailabilityTransformer(LoggableTransformer):

    def __init__(self, track_group_availability: DataFrame):
        """
        Filter any track group from the input dataset that is not available for streaming. This transformer is for
        non-personalized use-cases where we just verify that a track is stream ready in at least one country

        :param track_group_availability:  all available track groups (source: track group availability job)
        """
        super(TrackGroupAvailabilityTransformer, self).__init__()
        self.track_group_availability = track_group_availability

    def _transform(self, dataset) -> DataFrame:
        available_track_groups = self.track_group_availability.select(c.TRACK_GROUP)
        return dataset.join(available_track_groups, c.TRACK_GROUP)


def get_available_track_groups(tracks_metadata: DataFrame):
    """
    Get track groups that are available in at least one country

    :param tracks_metadata:     tracks metadata df
    """
    return (tracks_metadata
            .where(F.size(c.AVAILABLE_COUNTRY_CODES) > 0)
            .select(c.TRACK_GROUP)
            .distinct())
